module Quatroids

using Oscar
using GAP
using Plots
using Combinatorics

export
	#############Main Functions
	QuatroidAnalysis, #Runs all functions necessary to create results
	LoadMatroids, #Loads matroids if already parsed
	GenerateAllMatroids, #Loads, or computes, matroids
	AddConic, #Appends sextuple to Q, closes w.r.t. weak bezout and finds canonical rep
	AllConicExtensions, #Appends sextuples iteratively to Q=(I,[]) as in Alg 1
	LoadQuatroids, #Loads quatroids if already computed
	GenerateAllCandidateQuatroids, #Computes all candidate quatroids by iteratively applying Alg 1
	Bezoutian, #Computes boolean vector indicating if each candidate is Bezoutian
	OrbitSizes, #Computes integer vector of S_8 orbits of candidates
	ShortDescriptions, #Computes short descriptions of candidate quatroids
	QuatroidSignatures, #Computes identifiable signatures for each candidate
	QuatroidModifications, #Computes, for all Q with reduced base locus, all Q' such that Q~>Q'
	QuatroidReductions, #Reduces irreducibility of strata indexed by candidates to other strata
	QuatroidWeakUpperBounds, #Computes 12-m_Q, a weak upper bound for d_Q
	QuatroidStrongUpperBounds, #Computes the strong upper boudn for d_Q
	J2M2, #Converts julia syntax to Macaulay2 syntax

	############Auxiliary Functions

	#Helper Functions
	revLex, #The revlex order
	string_to_nonbases, #converts strings in SimpleMatroids38 file to nonbases
	OuterQuatroidLT, #An ordering for quatroids
	InnerQuatroidLT, #An ordering for triples and sextuples

	#GroupTheoreticFunctions
	minimalImage, #Computes a minimal image of Q under the S_8 action
	MatroidStabilizer, #Computes the subgroup of S_8 which stabilizes the matroidal part of Q
	MatroidStabilizedOrbit, #Computes the orbit of Q under the above stabilizer
	orb, #Computes the orbit of Q
	orbsize, #Computes the size of an orbit of Q

	#Quatroid Properties
	Triples, #Pulls triples from Q
	Sextuples, #Pulls sextuples from Q
	BezoutWeak1, #Determines if Q satisfies Bezouts first weak crit.
	BezoutWeak2, #Determines if Q satisfies Bezouts second weak crit.
	BezoutWeak3, #Determines if Q satisfies Bezouts third weak crit.
	BezoutWeak,  #Determines if Q satisfies Bezouts weak crit.
	BezoutWeakClosure1, #Closes triples w.r.t. Bezout's second weak crit.
	BezoutWeakClosure2, #Closes sextuples w.r.t. Bezout's third weak crit.
	BezoutClosure, #Closes triples and sextuples w.r.t. Bezout's second/third weak crit.
	BezoutStrong1, #Determines if Q satisfies Bezouts first strong crit.
	BezoutStrong2, #Determines if Q satisfies Bezouts second strong crit.
	BezoutStrong,  #Determines if Q satisfies Bezouts strong crit
	isBezoutian,   #Determines if Q satisfies Bezouts weak and strong crit.
	MatroidNumber, #Determines the index of the (orbit of the) underlying matroid of Q
	QuatroidSignature, #Produces a signature unique to the orbit of the quatroid
	QuatroidNumber, #Determines the index of the (orbit of the) underlying quatroid of Q
	shortDescription, #Returns the 'line flats' and 'conic flats' of Q
	ShortToQuatroid, #Turns the above description back into the quatroid format
	isExhaustive, #Checks if every point is involved in some condition
	ContainedIn10, #Checks if Q has a triple meeting a sextuple at one point
	ContainedIn77, #Checks if Q has two triples meeting at a point and the triple of that union's complement
	ReducedBaseLocus, #Checks if Q is exhaustive and isn't contained in Q_10 or Q_77
	mQ, #Computes the sum of multiplicities of reducible cubics in the pencil L_p for p representing Q

	#Irreducibility
	Modifications, #Computes all Q' such that Q~>Q' when Q has reduced base locus
	QuatroidReduction, #Computes a reduction of Q to Q' along with its reason
	ParseReductionData, #For each Q, iteratively reduces using i,ii,iii until it cannot anymore
	CompleteReductionParse, #Subsequently tries to reduce using condition iv

	#JuliaToM2
	J2M2


include("GroupTheoreticFunctions.jl")
include("HelperFunctions.jl")
include("QuatroidProperties.jl")
include("Irreducibility.jl")
include("JuliaToM2.jl")

###########################This function will run everything##############

function QuatroidAnalysis(;verbose=true)
	s="~~~~~~~~~~~~~~~~~~~~~"
	println(s*" Parsing Matroids From DB "*s)
	AllMatroids=GenerateAllMatroids(;verbose=true)

	println(s*" Computing All Candidate Quatroids "*s)
	AllQuatroids = GenerateAllCandidateQuatroids(;verbose=true)

	println(s*" Determining Which Candidate Quatroids Are Bezoutian"*s)
	Bezoutian()


	println(s*" Computing Orbit Sizes Of All Candidate Quatroid "*s)
	OrbitSizes()

	println(s*" Computing Short Descriptions Of All Candidate Quatroids "*s)
	ShortDescriptions()

	println(s*" Computing Signatures Of All Candidate Quatroids "*s)
	QuatroidSignatures()

	println(s*" Computing All Modifications of All Candidate Quatroids "*s)
	QuatroidModifications(;verbose=true)

	println(s*" Computing Reductions For All Candidate Quatroids "*s)
	QuatroidReductions(;verbose=true)

	println(s*" Computing Weak Upper Bounds For d_Q For All Candidate Quatroids "*s)
	QuatroidWeakUpperBounds()

	println(s*" Computing Strong Upper Bounds For d_Q For All Candidate Quatroids "*s)
	QuatroidStrongUpperBounds()

	println(s*" Reading Rational Representatives For All Candidate Quatroids "*s)
	R = LoadRationalRepresentatives()

	println(s*" Converting Data to M2 Format "*s)
	J2M2(R;flag="Representatives")
	J2M2(AllQuatroids;flag="Quatroids")
end

##########################################################################
#####################Matroid parsing from DB  ############################
##########################################################################

#Loads matroids from previously parsed file
function LoadMatroids(;verbose=false)
    JuliaMatroidLocation = pwd()*"/src/Results/Matroids.txt"
    R = readlines(JuliaMatroidLocation)
    AllMatroids=[]
    for r in R
        push!(AllMatroids,eval(Meta.parse(r)))
    end
    return(AllMatroids)
end

#Returns list of matroids, either by loading matroids or parsing them
function GenerateAllMatroids(;verbose=false)

	GAP.Globals.LoadPackage(GapObj("images")); #This package has minimal_image in it

	if verbose
		println("Checking if matroids have been previously parsed")
	end
    JuliaMatroidLocation = pwd()*"/src/Results/Matroids.txt"
    if isfile(JuliaMatroidLocation)
		if verbose
			println("Pulling parsed matroids")
		end
		#if matroids have already been parsed into julia syntax
		#   load them
        M = LoadMatroids()
        return(M)
    end
	if verbose
		println("Matroids have yet to be parsed. Parsing them now.")
	end
	try
    	minimalImage([[2,3,4]])
    catch e
           println("You must load GAP explicitly and run ")
		   println("""GAP.Globals.LoadPackage(GapObj("images"))""")
    end
    MatroidDatabaseLocation = pwd()*"/src/InitialFiles/SimpleMatroids38.txt"
	if isdir(pwd()*"/src/Results")==false
		mkdir(pwd()*"/src/Results")
	end
    if isfile(MatroidDatabaseLocation)
		#Read from the matroid database file
		R = readlines(MatroidDatabaseLocation)[1:68];

		#Convert the 0/* strings to sets of nonbases
		S = sort([string_to_nonbases(s,3,8) for s in R]);

		#Compute the minimal representative of each set of nonbases
		#  and sort the results
		AllMatroids = [minimalImage(s) for s in S];
		AllMatroids = sort([sort(s) for s in AllMatroids]);

		#Partition the resulting matroids into those which have
		# no four points on a line (Bezoutian) and those which do
		BezoutianMatroids = filter(x->issubset([[1,2,4]],x)==false,AllMatroids);
		NonBezoutianMatroids = filter(x->issubset([[1,2,4]],x),AllMatroids);

		#Reconstruct the list, placing those matroids with 4 points on a line last
		AllMatroids = vcat(BezoutianMatroids,NonBezoutianMatroids);

		#Include the rank 2 matroid where all 8 points are on a line
		push!(AllMatroids,sort(collect(combinations(1:8,3))));

		#numbers 8 and 50 are to be deleted:
		# 8 - The Fano matroid with a free eigth point
		# 50 - The Fano matroid with an eigth point on one of the lines
		deleteat!(AllMatroids,[x==8 || x==50 for x in 1:length(AllMatroids)]);

		touch(JuliaMatroidLocation)
		io = open(JuliaMatroidLocation,"w");
		write(io,"[]\n")
		for M in AllMatroids[2:end]
			write(io,string(M))
			write(io,"\n")
		end
		close(io)
		if verbose
			println("All matroids have been stored in the Results folder as Matroids.txt")
		end
		return(AllMatroids)
	end
end

##########################################################################
#################Generating Candidate Quatroids  #########################
##########################################################################

function AddConic(S,HS)
    #Construct a set of all possible ways to append
    # Dont forget to close the quatroids up
    SPlusConic=[BezoutClosure(vcat(S,[Q])) for Q in combinations(1:8,6)]
    # Take the distinct ones
    unique!(SPlusConic)
    # Check if they satisfy bezouts weak criterion 1 (no triple in a sextuple)
    filter!(x->BezoutWeak1(x),SPlusConic)
    # Find canonical representatives of the remaining potential quatroids
    SPlusConic = [MatroidStabilizedOrbit(q,HS) for q in SPlusConic]
    unique!(SPlusConic)
    return(SPlusConic)
end

function AllConicExtensions(Q;verbose=false)
	HS = MatroidStabilizer(Q)
	QuatroidsFound = [Q]
	QuatroidsToExtend = [Q]

	while length(QuatroidsToExtend)>0
		if verbose
			print("Number of quatroids found: ",length(QuatroidsFound))
			print("     Number of quatroids to extend: ",length(QuatroidsToExtend))
		end
		Extendable = popfirst!(QuatroidsToExtend)
		Extensions = AddConic(Extendable, HS)
		NewExtensions = filter(x->x∉QuatroidsFound,Extensions)
		if verbose
			print("     Number of new quatroids found: ",length(NewExtensions))
			println()
		end
		for newQ in NewExtensions
			push!(QuatroidsToExtend,newQ)
			push!(QuatroidsFound,newQ)
		end
	end

	return(QuatroidsFound)
end

function LoadQuatroids()
	JuliaQuatroidLocation = pwd()*"/src/Results/CandidateQuatroids.txt"
	R = readlines(JuliaQuatroidLocation)
	AllQuatroids=[]
	for r in R
	    push!(AllQuatroids,eval(Meta.parse(r)))
	end
	return(AllQuatroids)
end

function GenerateAllCandidateQuatroids(;verbose=false)
	JuliaQuatroidLocation = pwd()*"/src/Results/CandidateQuatroids.txt"
	if isfile(JuliaQuatroidLocation)
	#if quatroids have already been computed
	#   load them
    	Q = LoadQuatroids()
		return(Q)
	end

	AllMatroids=GenerateAllMatroids(;verbose=verbose)

	AllQuatroids=[]
	for i in 1:length(AllMatroids)
		M = AllMatroids[i]
		if verbose
			println("----------------------Extending Matroid ",i,"--------------------")
		end
		ACE = AllConicExtensions(M;verbose=verbose)
		for a in ACE
			push!(AllQuatroids,a)
		end
	end
	if verbose
		println("Sorting each quatroid internally")
	end
	AllQuatroids = [sort(Q; lt = InnerQuatroidLT) for Q in AllQuatroids]
	if verbose
		println("Sorting the set of quatroids externally")
	end
	AllQuatroids = sort(AllQuatroids; lt = OuterQuatroidLT)

	touch(JuliaQuatroidLocation)
	io = open(JuliaQuatroidLocation,"w");
	write(io,"[]\n")
	for Q in AllQuatroids[2:end]
		sq = string(Q)
		sq = replace(sq, "Any"=>"")
		write(io,sq)
		write(io,"\n")
	end
	close(io)
	println("All candidate quatroids have been stored in the Results folder as CandidateQuatroids.txt")
	return(AllQuatroids)
end

#############################################################
############  Additional Functions ##########################
#############################################################
function Bezoutian()
	AllQuatroids=GenerateAllCandidateQuatroids()
	BezoutLocation = pwd()*"/src/Results/Bezoutian.txt"
	if isfile(BezoutLocation)
		R = readlines(BezoutLocation)
		BezoutData=[]
		for r in R
		    push!(BezoutData,eval(Meta.parse(r)))
		end
		return(BezoutData)
	end
	touch(BezoutLocation)
	io = open(BezoutLocation,"w");
	Bez=[true]
	write(io,"true")
	for M in AllQuatroids[2:end]
		write(io,"\n")
		iB= isBezoutian(M)
		push!(Bez,iB)
		write(io,string(isBezoutian(M)))
	end
	close(io)
	println("Boolean values describing which candidate quatroids satisfy Bezout's strong criteria are in Bezoutian.txt")
	return(Bez)
end

function OrbitSizes()
	AllQuatroids=GenerateAllCandidateQuatroids()
	OrbitSizeLocation = pwd()*"/src/Results/OrbitSizes.txt"
	if isfile(OrbitSizeLocation)
		R = readlines(OrbitSizeLocation)
		OrbitSizes=[]
		for r in R
		    push!(OrbitSizes,eval(Meta.parse(r)))
		end
		return(OrbitSizes)
	end
	touch(OrbitSizeLocation)
	io = open(OrbitSizeLocation,"w");
	OrbitSizes=[1]
	write(io,"1")
	for M in AllQuatroids[2:end]
		write(io,"\n")
		os = orbsize(M)
		push!(OrbitSizes,os)
		write(io,string(orbsize(M)))
	end
	close(io)
	println("Sizes of orbits of quatroids are stored in OrbitSizes.txt")
	return(OrbitSizes)
end

function ShortDescriptions()
	AllQuatroids=GenerateAllCandidateQuatroids()
	SDLocation = pwd()*"/src/Results/ShortDescriptions.txt"
	if isfile(SDLocation)
		R = readlines(SDLocation)
		SD=[]
		for r in R
		    push!(SD,eval(Meta.parse(r)))
		end
		return(SD)
	end
	SD=[[[],[]]]
	touch(SDLocation)
	io = open(SDLocation,"w");
	write(io,"[[],[]]")
	for M in AllQuatroids[2:end]
		write(io,"\n")
		sd = shortDescription(M)
		sds = string(sd)
		sds = replace(sds, "Vector{Int64}"=>"")
		sds = replace(sds, "Vector{Any}"=>"")
		push!(SD,sd)
		write(io,string(sds))
	end
	close(io)
	println("Short descriptions of quatroids are stored in ShortDescriptions.txt")
	return(SD)
end

function QuatroidSignatures()
	AllQuatroids = GenerateAllCandidateQuatroids()

	SignatureLocation = pwd()*"/src/Results/QuatroidSignatures.txt"
	if isfile(SignatureLocation)
		R = readlines(SignatureLocation)
		Signatures=[]
		for r in R
			push!(Signatures,eval(Meta.parse(r)))
		end
		return(Signatures)
	end
	touch(SignatureLocation)
	io = open(SignatureLocation,"w");
	Signatures=[QuatroidSignature(AllQuatroids[1])]
	s1 = string(Signatures[1])
	s1 = replace(s1,"Any"=>"")
	write(io,s1)
	for Q in AllQuatroids[2:end]
		write(io,"\n")
		qs = QuatroidSignature(Q)
		qss = string(qs)
		qss = replace(qss, "Any"=>"")
		push!(Signatures,qs)
		write(io,qss)
	end
	close(io)
	println("Quatroid signatures are stored in Signatures.txt")
	return(Signatures)
end


function QuatroidModifications(;verbose=false)
	ModificationLoc = pwd()*"/src/Results/Modifications.txt"
	if isfile(ModificationLoc)
		R = readlines(ModificationLoc)
		Mods=[]
		for r in R
			push!(Mods,eval(Meta.parse(r)))
		end
		return(Mods)
	end
	AllQuatroids = GenerateAllCandidateQuatroids()
	Mods=[[1]]
	touch(ModificationLoc)
	io = open(ModificationLoc,"w");
	write(io,string(Mods[1]))
	for i in 2:126
		if verbose
			println("Modifications of quatroid candidate ",i,":")
		end
		Q = AllQuatroids[i]
		write(io,"\n")
		Mod=[]
		if i!=63
			Mod = Modifications(Q)
		else
			Mod = [63]
		end
		if verbose
			println("   ",Mod)
		end
		push!(Mods,Mod)
		write(io,string(Mod))
	end
	close(io)
	println("All quatroid modifications are stored in Modifications.txt")
	return(Mods)
end

function QuatroidReductions(;verbose=false)
	ReductionDictionary = Dict{Int64,Any}()
	ReasonDictionary = Dict{Int64,Any}()

	for i in 1:126
		if verbose
			println("Reducing quatroid ",i)
		end
		Result = QuatroidReduction(i)
		Reason = Result[1]
		ReducedTo = Result[2]
		if verbose
			println("Quatroid ",i," reduces to quatroid ",ReducedTo," for reason ",Reason)
		end
		ReductionDictionary[i] = ReducedTo
		ReasonDictionary[i] = Reason
	end
	PRD = ParseReductionData(ReductionDictionary,ReasonDictionary;verbose=verbose)
	Reductions=PRD[1]
	ReductionProofs=PRD[2]

	ReductionLoc = pwd()*"/src/Results/Reductions.txt"
	ReductionProofLoc = pwd()*"/src/Results/ReductionProofs.txt"

	touch(ReductionLoc)
	io = open(ReductionLoc,"w");
	write(io,string(Reductions[1]))
	for i in 2:126
		write(io,"\n")
		write(io,string(Reductions[i]))
	end
	close(io)
	println("Reductions are written in Reductions.txt")


	touch(ReductionProofLoc)
	io = open(ReductionProofLoc,"w");
	write(io,ReductionProofs[1])
	for i in 2:126
		write(io,"\n")
		write(io,ReductionProofs[i])
	end
	close(io)
	println("Reduction proofs are written in ReductionProofs.txt")

	return(PRD)
end

function QuatroidWeakUpperBounds(;verbose=false)
	WeakUpperBounds=[]
	AllQuatroids = GenerateAllCandidateQuatroids()
	for i in 1:126
		Q=AllQuatroids[i]
		if i==63 || isBezoutian(Q)==false
			push!(WeakUpperBounds,0)
		else
			m=mQ(Q)
			push!(WeakUpperBounds,12-m)
		end
	end
	WUBloc = pwd()*"/src/Results/WeakUpperBounds.txt"
	touch(WUBloc)
	io = open(WUBloc,"w");
	write(io,string(WeakUpperBounds[1]))
	for i in 2:126
		write(io,"\n")
		write(io,string(WeakUpperBounds[i]))
	end
	close(io)
	println("Weak upper bounds for dQ are stored in WeakUpperBounds.txt")

	return(WeakUpperBounds)
end

function QuatroidStrongUpperBounds(;verbose=false)
	StrongUpperBounds=[]
	AllQuatroids = GenerateAllCandidateQuatroids()
	for i in 1:126
		Q=AllQuatroids[i]
		if i==63 || isBezoutian(Q)==false
			push!(StrongUpperBounds,0)
		else
			m=mQ(Q)
			if ContainedIn10(Q) || ContainedIn77(Q)
				m=m+1
			end
			push!(StrongUpperBounds,12-m)
		end
	end

	SUBloc = pwd()*"/src/Results/StrongUpperBounds.txt"
	touch(SUBloc)
	io = open(SUBloc,"w");
	write(io,string(StrongUpperBounds[1]))
	for i in 2:126
		write(io,"\n")
		write(io,string(StrongUpperBounds[i]))
	end
	close(io)
	println("Strong upper bounds for dQ are stored in StrongUpperBounds.txt")

	return(StrongUpperBounds)
end

function LoadRationalRepresentatives()
	RatlRepLoc = pwd()*"/src/InitialFiles/RationalRepresentatives.txt"
	R = readlines(RatlRepLoc)
	RationalRepresentatives=[]
	for r in R
		M = eval(Meta.parse(r))
		if M!=[]
	    	push!(RationalRepresentatives,M)
		else
			push!(RationalRepresentatives,M)
		end
	end
	return(RationalRepresentatives)
end

end # module Quatroids
